/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Dict
 *
 * The Initial Developer of the Original Code is David Costanzo
 * Portions created by the Initial Developer are Copyright (C) 2003-2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// dictInit will be called every time a new browser window is created
window.addEventListener("load", dictInit, false);


function utf8ToCharacterEntitiesInternal(utf8String) {

    var encodedString = '';
    var octetsRemainingInCharacterCode = 0;
    var characterCode = 0;

    for (var i = 0; i < utf8String.length; i++) {

        var inputCharCode = utf8String.charCodeAt(i);

        if ((inputCharCode & 0x80) == 0x00) {
            // this can be encoded with ASCII-7
            encodedString += utf8String[i];

            // reset the multi-octet stream accumulator.
            // if it was not zero, then the input was bad.
            octetsRemainingInCharacterCode = 0;
        }
        else if ((inputCharCode & 0xC0) == 0x80) {
            // 10xxxxxx

            // this is a continuation of a multi-octet sequence
            characterCode = (characterCode << 6) | (inputCharCode & 0x3F);

            octetsRemainingInCharacterCode--;
            if (octetsRemainingInCharacterCode == 0) {
                // this is the final octet in the mutli-octet code
                encodedString += "&#" + characterCode + ";";
            }
            else if (octetsRemainingInCharacterCode < 0) {
                // bad input
                octetsRemainingInCharacterCode = 0;
            }
        }
        else if ((inputCharCode & 0xE0) == 0xC0) {
            // 110xxxxx 10xxxxxx
            characterCode = inputCharCode & 0x1F;
            octetsRemainingInCharacterCode = 1;
        }
        else if ((inputCharCode & 0xF0) == 0xE0) {
            // 1110xxxx 10xxxxxx 10xxxxxx
            characterCode = inputCharCode & 0x0F;
            octetsRemainingInCharacterCode = 2;
        }
        else if ((inputCharCode & 0xF8) == 0xF0) {
            // 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
            characterCode = inputCharCode & 0x07;
            octetsRemainingInCharacterCode = 3;
        }
        else if ((inputCharCode & 0xFC) == 0xF8) {
            // 111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
            characterCode = inputCharCode & 0x03;
            octetsRemainingInCharacterCode = 4;
        }
        else if ((inputCharCode & 0xFE) == 0xFC) {
            // 1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
            characterCode = inputCharCode & 0x01;
            octetsRemainingInCharacterCode = 5;
        }
        else {
            // Illegal UTF-8 code.  Drop it.
        }
    }

    return encodedString;
}

//
// utf8ToCharacterEntities() converts an octet stream that is 
// encoded in UTF-8 into an ASCII-7 encoding of the same 
// stream such that all character > 127 are encoded using
// HTML character entities &#xxx;.
//
// This is necessary when using international DICT servers,
// because I cannot figure out another way to tell Mozilla
// to interpret the string as UTF-8.
//
//   UCS-4 range (hex.)    UTF-8 octet sequence (binary)
//   0000 0000-0000 007F   0xxxxxxx
//   0000 0080-0000 07FF   110xxxxx 10xxxxxx
//   0000 0800-0000 FFFF   1110xxxx 10xxxxxx 10xxxxxx
//   0001 0000-001F FFFF   11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
//   0020 0000-03FF FFFF   111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
//   0400 0000-7FFF FFFF   1111110x 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx
//
function utf8ToCharacterEntities(utf8String) {

    // Before we do any convertion, let's see if there are
    // any character codes outside of the ASCII-7 range.
    // If not, then the string doesn't need to be converted.

    if (!utf8String.match(/[\x80-\xFF]/)) {
        return utf8String;
    }

    // Note that we could just call utf8ToCharacterEntitiesInternal()
    // on the entire string, but this routine is very slow.
    // Instead, it is faster to use a regular expression
    // to separate the octet-sequences that must be replaced
    // while preserving long runs of characters that don't to escaping.

    // This makes the routine run in O(n^2) of the characters that
    // must be replaced, instead of O(n^2) of all characters.
    // When the text is mostly ASCII-7, this is a significant savings.

    var encodedString = '';
    
    var matchExp = /([\x00-\x7F]*)([\x80-\xFF]+)([\x00-\x7F]*)/g;
    while ((result = matchExp.exec(utf8String)) != null) {

        var singleOctetCodesBefore = result[1];
        var multiOctetCodes        = result[2];
        var singleOctetCodesAfter  = result[3];
        
        encodedString += 
            singleOctetCodesBefore + 
            utf8ToCharacterEntitiesInternal(multiOctetCodes) +
            singleOctetCodesAfter;
    }

    return encodedString;
}


// DictListener class, which is passed in as the "listener" parameter
// to DictClient::defineWord().
function DictListener () {
}

// wraps some HTML inside HTML/BODY tag and adds a title
DictListener.prototype = 
{

    createHtmlDocument : function (title, body)
    {
        var htmlDocument;
    
        htmlDocument = 
            '<HTML>\n'    +
            '  <HEAD>\n'  + 
            '      <TITLE>' + utf8ToCharacterEntities(title) + '</TITLE>\n' +
            '  </HEAD>\n' +
            '  <BODY>\n'  +
            utf8ToCharacterEntities(body) +
            '  </BODY>\n' +
            '</HTML>\n';
        
        return htmlDocument;
    },

    // Creates a new window and sets the document to be the given HTML.
    // Resizes the window to fit the content.
    createWindow : function (htmlDocument)
    {
        const features = [
            'status=no', 
            'toolbar=no',
            'menubar=no',
            'location=no',
            'width=450',
            'height=100'
        ];

        var popupWindow = window.open(
            '',
            '',
            features.join(','));
    
        popupWindow.document.clear();
        popupWindow.document.write(htmlDocument);
        popupWindow.document.close();
    
        // HACK: Figure out how to size the window
        //       to fit the content
        popupWindow.sizeToContent();
        popupWindow.innerHeight += 70;
        
        // If it's too large to display without scrollbars, shorten it
        if (600 < popupWindow.innerHeight) {
            // limit the height
            popupWindow.innerHeight = 600;
            
            // make wider to account for the width of the 
            // vertical scrollbar.  If we don't do this, a 
            // horizontal scrollbar will appear.
            popupWindow.innerWidth += 20;  
        }

        return popupWindow;
    },

    //
    // The next three functions are the DictListener interface
    //

    // Called when the DictClient has content to render
    // (either definitions or closest matches)
    onHaveDocument : function (title, body)
    {
        var htmlDocument = this.createHtmlDocument(title, body);
        
        this.createWindow(htmlDocument);
    },

    onNoDefinitionsFound : function (word) 
    {
        alert('No definitions were found for "' + word + '".\n');
    },
    
    onError : function (error)
    {
        alert(error);
    }
}



// Returns the selected text as a string.
function getSelection() 
{
    var selectedWord = gContextMenu.searchSelected();

    // sanitize the selection
    selectedWord = selectedWord.replace(/^\W+/, ''); // trim leading non-alnum chars
    selectedWord = selectedWord.replace(/\W+$/, ''); // trim trailing non-alnum chars
    selectedWord = selectedWord.replace(/\s+/, ' '); // conflate whitespace runs

    return selectedWord;
}

        
function defineSelection()
{
    var selectedWord = getSelection();

    if (selectedWord.indexOf('"') != -1) {
        alert("Sorry, selections that contain quotation marks cannot be defined.");
        return;
    }

    if (selectedWord.length == 0) {
        alert("Please select a word to define.");
        return;
    }

    try {
        var pref = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch);

        // read the host and port from the preferences
        var host;
        var port;

        try {
            host = pref.getCharPref('extensions.dict.defaultservername');
            port = pref.getIntPref('extensions.dict.defaultserverport');
        }
        catch (e) {
            // REVIST: 
            //   If there the preferences are not set/corrupt, should 
            //   we just use a reasonable default or should we warn the
            //   user.
            host = 'dict.org';
            port = 2628;
        }

        var listener = new DictListener();
        var client   = new DictClient();
        client.defineWord(host, port, selectedWord, listener);
    }
    catch (ex) {
        alert(ex);
    }
}


function dictInit() {

    var contextMenu = document.getElementById("contentAreaContextMenu");

    // The first time dictInit is called, contextMenu is null.
    // This test prevents an error from being logged in the JavaScript Console
    if (contextMenu != null) {
        contextMenu.addEventListener(
            "popupshowing",
            onDictPopup,
            false);  
    }
}




// hide the Dictionary Search item when apropriate (use same logic as for web search)
function onDictPopup() {

    var item = document.getElementById('dict-define-selection');
    if (gContextMenu.isTextSelected) {

        var selection = getSelection();
        
        const LENGTHCUTOFF = 15;

        if (LENGTHCUTOFF < selection.length) {
            // selection is too long to appear in menu.  It must be abbreviated.
            selection = selection.slice(0, LENGTHCUTOFF) + '...';
        }

        var dictStringBundle = document.getElementById("dict_string_bundle");
        var labelTemplate = dictStringBundle.getString("context_searchselect.label");

        item.setAttribute("label", labelTemplate.replace('%S', selection));
        item.hidden = false; // show the menu item
    }
    else {
        item.hidden = true; // hide the menu item
    }
}
